#include <fvsp.h>
#include <string.h>
#include <stdlib.h>

struct part {
    char *text;
    size_t length; /* Excluding trailing '\0' */
    struct part *next;
};


/* Deallocate list of parts recuresively */
static void part_free(struct part *part) {
    while (part != NULL) {
        struct part *next = part->next;
        free(part->text);
        free(part);
        part = next;
    }
}


/* Create a part with a string from @start inclusively to @end exclusively */
static struct part *part_new(const char *start, const char *end) {
    struct part *part = NULL;

    part = malloc(sizeof(*part));
    if (NULL == part) return NULL;

    part->length = end - start;
    part->next = NULL;

    part->text = malloc(part->length + 2 + 1); /* +2 for padding */
    if (part->text == NULL) {
        part_free(part);
        return NULL;
    }
    memcpy(part->text, start, part->length);
    part->text[part->length] = 0;

    return part;
}


/* Append @source text of @source_length bytes to @output at @output_length
 * offset and add @source_length to @length number. The lengths does not count
 * traling '\0'. */
static void append(char *output, size_t *output_length,
        const char *source, size_t source_length) {
    memcpy(output + *output_length, source, source_length);
    *output_length += source_length;
}


/* Convert a Perl version value into RPM version string format.
 * @input is Perl version string, fraction or dotted one. Empty string is
 * equaled to "0".
 * @shorten is true to cut trailing zero groups. If no digit would remain,
 * empty string will be returned.
 * @return RPM-compatible version string or NULL in case of error. Caller is
 * responsible for deallocating the value by free(3). */
char *vsp(const char *input, _Bool shorten) {
    char *output = NULL;
    size_t output_length;
    const char *input_end, *input_start;
    struct part *parts = NULL;
    struct part *last_part;
    _Bool is_vstring;
    const char *zero = "0";

    if (NULL == input) {
        return NULL;
    }
    if ('\0' == *input) {
        input = zero;
    }

    /* Remove underscore parts */
    for (input_end = input; *input_end != '\0' && *input_end != '_';
            input_end++);
    
    if (input == input_end) {
        input = zero;
        input_end = zero + 1;
    }

    /* Dot is a delimiter. Split to parts. */
    input_start = input;
    for (const char *c = input; c < input_end; c++) {
        if (*c == '.' || c + 1 == input_end) {
            struct part *part = part_new(input_start,
                    (*c == '.') ? c : input_end);
            if (part == NULL) {
                part_free(parts);
                return NULL;
            }
            if (NULL == parts) parts = part;
            else last_part->next = part;
            last_part = part;

            input_start = c + 1;

            /* Convert trailing dot to a new "0" part */
            if (*c == '.' && c + 1 == input_end) {
                struct part *part = part_new(zero, zero + 1);
                if (part == NULL) {
                    part_free(parts);
                    return NULL;
                }
                if (NULL == parts) parts = part;
                else last_part->next = part;
                last_part = part;
            }
        }
    }
    /* Normalize leading part */
    if (parts->length == 0) {
        append(parts->text, &parts->length, zero, 1);
    }

    /* Preserve leading part */
    output = malloc((input_end - input) * 4 + 4 + 1); /* *4 for padding
                                             +4 for minimal ".0.0" suffix */
    if (NULL == output) {
        part_free(parts);
        return NULL;
    }
    output_length = 0;
    /* Is this version string or fraction string? */
    if (*parts->text == 'v') {
        append(output, &output_length, parts->text + 1, parts->length- 1);
        is_vstring = 1;
    } else {
        append(output, &output_length, parts->text, parts->length);  
        if (NULL != parts->next && NULL != parts->next->next) {
            is_vstring = 1;
        } else {
            is_vstring = 0;
        }
    }
    /* Discard leading part */
    last_part = parts->next;
    parts->next = NULL;
    part_free(parts);
    parts = last_part;

    /* Reformat parts after leading dot */
    if (!is_vstring && parts != NULL) {
        /* If it's not a vstring and there is only one part after leading
         * dot, it's a fraction number. */
        /* ???: We could raise an error if there were more parts. That's
         * invalid version string refused by Perl. */
        size_t trailer, length;
        char *digits = NULL; 

        /* Augment digits to factor of 3 */
        trailer = parts->length % 3;
        if (trailer) {
            append(parts->text, &parts->length, zero, 1);
            if (trailer == 1) {
                append(parts->text, &parts->length, zero, 1);
            }
        }
        digits = parts->text;
        length = parts->length;
        parts->text = NULL;
        part_free(parts);
        parts = NULL;

        /* Split it into triples */
        for (size_t offset = 0; offset < length; offset += 3) {
            struct part *part = part_new(digits + offset, digits + offset + 3);
            if (part == NULL) {
                part_free(parts);
                free(digits);
                free(output);
                return NULL;
            }
            if (NULL == parts) parts = part;
            else last_part->next = part;
            last_part = part;
        }

        free(digits);
    }

    /* Append necessary number of parts to get X.Y.Z format */
    if (parts == NULL) {
        parts = part_new(zero, zero + 1);
        if (parts == NULL) {
            free(output);
            return NULL;
        }
    }
    if (parts->next == NULL) {
        struct part *part = part_new(zero, zero + 1);
        if (part == NULL) {
            part_free(parts);
            free(output);
            return NULL;
        }
        parts->next = part;
    }

    /* Concatenate parts */
    for (const struct part *part = parts; part != NULL; part = part->next) {
        /* Strip leading zeros */
        size_t offset = 0;
        while (offset < part->length && part->text[offset] == '0') {
            offset++;
        }
        if (offset == part->length) offset--;
        append(output, &output_length, ".", 1);
        append(output, &output_length, part->text + offset,
                part->length - offset);
    }

    /* Clean up */
    part_free(parts);

    /* Shorten if requested */
    if (shorten && output_length > 0) {
        size_t offset, cut_at;

        /* Cut off all trailing zero groups and drop 0 integer */
        offset = output_length - 1;
        cut_at = output_length;
        while (output[offset] == '0' || output[offset] == '.') {
            if (output[offset] == '.')
                cut_at = offset;
            if (offset == 0) {
                cut_at = 0;
                break;
            }
            offset--;
        }
        if (cut_at != output_length) {
            output_length = cut_at;
        }
    }

    /* Terminate output string properly */
    output[output_length] = '\0';
    return output;
}
